/*
** Module   : AppServerLauncher.java
** Abstract : Class dedicated to launching appservers.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130705 Created initial version.
** 002 CA  20140124 Mark the client to be launched batch mode. Redirect the standard output and
**                  error streams of the client process to a file.
** 003 CA  20140206 Changed to launch the appserver using the spawner and scheduler infrastructure.
**                  Moved logging support to use the logging provided by LogHelper. 
** 004 MAG 20140310 When running on Windows OS we have to provide a password as well.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * This is a launcher of appservers and is used in two ways:
 * <ul>
 *    <li>Automatically starts the appservers which have the <code>auto_start=true</code> in their
 *        configuration; this will start up to <code>initial_agents</code>.</li>
 *    <li>Allow starting of appservers remotely; if a server is started using the "-a appserver"
 *        argument, it will connect to the target P2J Server and call {@link AppServer#launch} to
 *        launch the specified appserver. If the appserver is already started, it will attempt to
 *        start an agent.</li>
 * </ul>
 * 
 * There is one other way to start an appserver: by manually specifying the ID of an account 
 * configured to start an appserver, when starting a P2J Client.  This will start only one Agent
 * (if non started).
 */
public final class AppServerLauncher
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(AppServerDefinition.class.getName());

   /** The maximum number of continous failures to start an Agent, before giving up. */
   private static final int MAX_FAILURES = 10;
   
   /** The maximum amount of time to wait for an Agent to start. */
   private static final int LAUNCH_TIMEOUT = 120000;
   
   /** Map of launchers, per appserver name. */
   private static final Map<String, Launcher> launchers = new HashMap<String, Launcher>();

   /**
    * Launch the specified appserver. This includes starting the specified number of initial 
    * agents.
    * 
    * @param    appServer
    *           The appserver name to launch.
    * 
    * @return   <code>true</code> if the appserver was launched.
    */
   public static boolean launch(String appServer)
   {
      SecurityManager secMgr = SecurityManager.getInstance();
   
      // find the user of this appserver
      String[] subjects = secMgr.getAccountsForAppserver(appServer);
      if (subjects.length == 0)
      {
         final String msg = "Appserver '%s' has no associated user.";
         throw new IllegalStateException(String.format(msg, appServer));
      }
      else if (subjects.length > 1)
      {
         final String msg = "Appserver '%s' is associated with more than one user: %s";
         throw new IllegalStateException(String.format(msg, 
                                                       appServer,
                                                       Arrays.asList(subjects)));
      }
      
      return launch(subjects[0], appServer);
   }

   /**
    * When the appserver Agent is actually about to start listening for commands, it will notify 
    * its launcher that it was started. Until then, the appserver Agent is presumed "not running".
    * <p>
    * Note that when an appserver is configured with <code>auto_start=false</code> and the agent
    * is started manually, no {@link Launcher} will be available.  Launchers are used only the 
    * appserver is started automatically or when the server receives the "-p appserver" argument.
    * 
    * @param    appServer
    *           The name of the appserver with which the Agent is associated.
    * @param    success
    *           <code>true</code> if the Agent was started.
    */
   static void notifyLaunch(String appServer, boolean success)
   {
      Launcher launcher = null;
      synchronized (launchers)
      {
         launcher = launchers.get(appServer);
      }
      
      // the launcher can be null if started manually
      if (launcher != null)
      {
         synchronized (launcher)
         {
            launcher.success = success;
   
            launcher.notify();
         }
      }
   }
   
   /**
    * Launch an agent for the specified appserver.  The agent will need to authenticate using the
    * specified account.
    * <p>
    * If no agent is alive for this appserver, it will initialize the appserver by starting an
    * <code>initial_agents</code> Agents.  An Agent startup will fail if it takes more than 
    * {@value #LAUNCH_TIMEOUT} seconds.  Also, the Agent startup will not continue if it failed
    * more than {@link #MAX_FAILURES} times in a row.
    * <p>
    * If first call and at least one agent has been started, it will start a thread responsible
    * of sending the 
    * 
    * @param    account
    *           The account used to authenticate the Agent.
    * @param    appServer
    *           The appserver name.
    *           
    * @return   <code>true</code> if the appserver or agent was started successfully.
    */
   private static boolean launch(String account, final String appServer)
   {
      Launcher launcher = null;
      boolean first = false;
      synchronized (launchers)
      {
         launcher = launchers.get(appServer);
         if (launcher == null)
         {
            first = true;
            launcher = new Launcher(account);
            launchers.put(appServer, launcher);
         }
      }
      
      synchronized (launcher)
      {
         launcher.success = null;
      }
      
      AppServerDefinition appDef = AppServerDefinition.get(appServer);
      boolean success = true;
      if (first)
      {
         int failures = 0;
         int initialAgents = appDef.getInitialAgents();
         while (true)
         {
            boolean launched = launchWorker(appServer, launcher);
            failures = (launched ? 0 : failures + 1);
            
            AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
            if (pool != null && pool.size() >= initialAgents)
            {
               break;
            }
            
            if (failures >= MAX_FAILURES)
            {
               // protect this loop so it is finite; if agent startup failed 10 times in a row,
               // abend the process

               if (LOG.isLoggable(Level.WARNING))
               {
                  if (pool == null)
                  {
                     final String msg = "Could not start appserver '%s'.";
                     LOG.logp(Level.WARNING, "AppServerLauncher", "launch",
                              String.format(msg, appServer));
                     success = false;
                  }
                  else
                  {
                     final String msg = "Only %d out of %d were started for appserver %s";
                     LOG.logp(Level.WARNING, "AppServerLauncher", "launch",
                              String.format(msg, pool.size(), initialAgents, appServer));
                  }
               }
               
               break;
            }
         }
         
         // start the trim-agents timer
         if (success)
         {
            synchronized (launcher)
            {
               if (launcher.timer == null)
               {
                  launcher.timer = new Timer(appServer + " agent trimming thread.", true);
                  TimerTask tt = new TimerTask()
                  {
                     public void run()
                     {
                        trimAgents(appServer);
                     }
                  };
                  long timeout = appDef.getAutoTrimTimeout() * 1000;
                  launcher.timer.scheduleAtFixedRate(tt, timeout, timeout);
               }
            }
         }
      }
      else
      {
         success = launchWorker(appServer, launcher);
      }
      
      return success;
   }

   /**
    * Terminate the given appserver.
    * 
    * @param    appServer
    *           The name of the appserver to terminate.
    */
   public static void terminate(String appServer)
   {
      AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
      pool.terminateAll();
   }
   
   /**
    * Trim the agents for the specific appserver down to a <code>min_agents</code> count.
    * 
    * @param    appServer
    *           The appserver name.
    */
   public static void trimAgents(String appServer)
   {
      AgentPool pool = AppServerManager.getAgentPoolForAppserver(appServer);
      pool.trimAgents();
   }

   /**
    * Notify that the given appserver has been terminated (there are no more agents for it and
    * the pool is shutting down).
    * 
    * @param    appServer
    *           The name of the appserver being terminated.
    */
   static void terminated(String appServer)
   {
      Launcher launcher = null;
      synchronized (launchers)
      {
         launcher = launchers.remove(appServer);
      }
      
      if (launcher != null)
      {
         synchronized (launcher)
         {
            launcher.timer.cancel();
         }
      }
   
      AppServerDefinition.remove(appServer);
   }

   /**
    * Worker for Agent launching.  After starting the Agent via the {@link Launcher#getSpawner()},
    * it will wait for {@link #LAUNCH_TIMEOUT} seconds for the Agent to call {@link #notifyLaunch}.
    * <p>
    * After the timeout  or when it receives the launch notification, it will check the 
    * {@link Launcher#success} flag, for the status of the launch.  Note that the {@link Launcher}
    * is shared among processes trying to launch Agents; thus, the launch notification may come
    * from a different Agent than the one currently starting.  This is not a problem, the party
    * starting the Agent doesn't care which agent started, it will only need to know if an Agent
    * was started; other parties waiting for Agent startup will eventually pick up the launch
    * notification, if the Agent was started.
    * 
    * @param    appServer
    *           The name of the appserver to be launched.
    * @param    launcher
    *           The launcher used to start the agent.
    *           
    * @return   <code>true</code> if the Agent was started successfully.
    */
   private static boolean launchWorker(String appServer, Launcher launcher)
   {
      boolean success = false;
      synchronized (launcher)
      {
         synchronized (launchers)
         {
            ClientSpawner spawner = launcher.getSpawner();
            
            SecurityManager sm = SecurityManager.getInstance();
            
            // get the user name from the directory
            spawner.spawn(sm.getSystemUser(launcher.account), 
                          sm.getSystemPassword(launcher.account));
         }
         
         try
         {
            launcher.wait(LAUNCH_TIMEOUT);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
         
         success = (launcher.success != null && launcher.success);

         // reset the value, so is not seen by other threads
         launcher.success = null;
      }
      
      if (LOG.isLoggable(Level.INFO))
      {
         final String msgSuccess = "Appserver '%s' was started sucessfully!";
         final String msgFailure = "Appserver '%s' could not be started!";
         LOG.logp(Level.INFO, "AppServerLauncher", "launchWorker", 
                  String.format(success ? msgSuccess : msgFailure, appServer));
      }
      
      return success;
   }
   
   /**
    * Container for a Agent launching information.
    */
   private static class Launcher
   {
      /** The P2J account to which the appserver is registered. */
      private final String account;

      /**
       * When set to {@link Boolean#TRUE}, it will indicate the Agent was started successfully;
       * any other value will indicate failure.
       */
      private Boolean success;

      /** Timer for trimming the Agents. */
      private Timer timer;
      
      /**
       * Create a new Agent launcher.
       * 
       * @param    account
       *           The P2J account to which the appserver is registered.
       */
      private Launcher(String account)
      {
         this.account = account;
      }
      
      /**
       * Get a new spawner used to launch the appserver process.
       * 
       * @return   See above.
       */
      public ClientSpawner getSpawner()
      {
         return new ProcessClientSpawner(account);
      }
   }
}
